<?php
/*
 * Copyright 2007-2017 Abstrium <contact (at) pydio.com>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com>.
 */
namespace Pydio\Core\Http\Client;

use DateTime;
use Exception;
use GuzzleHttp\Client;
use Pydio\Core\Exception\LoginException;
use Pydio\Core\Services\ConfService;
use Pydio\Log\Core\Logger;
use Swagger\Client\Api\TokenServiceApi;
use Swagger\Client\Model\RestRevokeRequest;

/**
 * Class DexApi
 * @package Pydio\Core\Http\Client
 */
class DexApi extends MicroApi {

    protected $dexUrl;
    protected $clientID;
    protected $clientSecret;

    public function __construct()
    {
        parent::__construct();
        $this->dexUrl = ConfService::getGlobalConf("ENDPOINT_DEX", "conf");
        $this->clientID = ConfService::getGlobalConf("ENDPOINT_DEX_CLIENTID", "conf");
        $this->clientSecret = ConfService::getGlobalConf("ENDPOINT_DEX_CLIENTSECRET", "conf");
    }

    /**
     * @param $token
     * @return \stdClass
     */
    private function claimsFromString($token) {
        $b64urlb64 = function ($str) {
            $padding = strlen($str) % 4;
            if ($padding > 0) {
                $str .= str_repeat("=", 4 -$padding);
            }
            return strtr($str, '-_', '+/');
        };
        $parts = explode(".", $token);
        $claims = json_decode(base64_decode($b64urlb64($parts[1])));
        return $claims;
    }

    /**
     * @param $tokens
     * @return bool
     */
    public function isExpired($tokens) {
        $claims = $this->claimsFromString($tokens["id_token"]);
        $expire_in = intval($tokens["expires_in"]);
        $date = new DateTime();
        $date->setTimeStamp($claims->iat + $expire_in);
        return $date < new DateTime("now");
    }

    /**
     * @param $userId
     * @param $userPass
     * @return mixed|null
     * @throws LoginException
     */
    public function getToken($userId, $userPass) {

        $client = new Client();
        $response = $client->post($this->dexUrl . '/dex/token', [
            "body" => [
                "grant_type" => "password",
                "scope"      => "email profile pydio",
                "username"   => $userId,
                "password"   => $userPass,
                "nonce"      => uniqid(),
            ],
            "headers" => [
                "Authorization" => "Basic ". base64_encode("$this->clientID:$this->clientSecret"), // ClientID + Client Secret as defined is config-dev.yaml
            ]
        ]);
        $full = $response->getBody()->getContents();
        $json = json_decode($full, true);
        if ($json == null) {
            throw new \Pydio\Core\Exception\LoginException(-1);
        }
        return $json;
    }

    /**
     * @param array
     * @return array
     */
    public function refreshToken($refresh_token) {
        try{
            $client = new Client();
            $response = $client->post($this->dexUrl . '/dex/token', [
                "body" => [
                    "grant_type" => "refresh_token",
                    "scope"      => "email profile pydio offline_access",
                    "nonce"      => uniqid(),
                    "refresh_token" => $refresh_token
                ],
                "headers" => [
                    "Authorization" => "Basic ". base64_encode("$this->clientID:$this->clientSecret"), // ClientID + Client Secret as defined is config-dev.yaml
                ]
            ]);
            $full = $response->getBody()->getContents();
            $json = json_decode($full, true);
            return $json;
        }catch (Exception $e){
            Logger::error("Could not get DEX AUTH TOKEN, IGNORING", "", "");
        }
        return null;
    }

    /**
     * @return mixed
     */
    public function revokeToken($token) {

        $api = new TokenServiceApi(self::getApiClient());
        $request = new RestRevokeRequest();
        $request->setTokenId($token);
        $response = $api->revoke($request);
        return $response;

    }

}
