<?php
/*
 * Copyright 2007-2017 Abstrium <contact (at) pydio.com>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com/>.
 */
namespace Pydio\Core\Services;

use Pydio\Conf\Core\Role;
use Pydio\Core\Exception\PydioException;
use Pydio\Core\Http\Client\MicroApi;
use Pydio\Core\Model\RepositoryRoot;
use Pydio\Core\Model\UserInterface;
use Pydio\Core\PluginFramework\PluginsService;
use Swagger\Client\ApiException;
use Swagger\Client\Model\IdmACL;
use Swagger\Client\Model\IdmACLAction;
use Swagger\Client\Model\IdmACLSingleQuery;
use Swagger\Client\Model\IdmRoleSingleQuery;
use Swagger\Client\Model\RestSearchACLRequest;
use Swagger\Client\Model\RestSearchRoleRequest;
use Swagger\Client\Model\ServiceOperationType;

defined('PYDIO_EXEC') or die('Access not allowed');

/**
 * Class RolesService
 * @package Pydio\Core\Services
 */
class RolesService
{

    const RootGroup = "ROOT_GROUP";

    /**
     * @todo MOVE TO BACKEND
     * Update a user with admin rights and return it
    * @param UserInterface $adminUser
     * @return UserInterface
    */
    public static function updateAdminRights($adminUser)
    {
        if ($adminUser->getPersonalRole()->getAcl('settings') != "rw") {
            $adminUser->getPersonalRole()->setAcl('settings', 'rw');
            self::updateRole($adminUser->getPersonalRole(), $adminUser);
            $adminUser->recomputeMergedRole();
            $adminUser->save("superuser");
        }
        return $adminUser;
    }

    /**
     * Update a user object with the default repositories rights
     * @todo MOVE TO BACKEND
     * @param UserInterface $userObject
     */
    public static function updateDefaultRights(&$userObject)
    {
        if (!$userObject->hasParent()) {
            $rolesList = self::getRolesList(array(), true);
            foreach ($rolesList as $roleId => $roleObject) {
                if (!$userObject->canSee($roleObject)) continue;
                if ($userObject->getProfile() == "shared" && $roleObject->autoAppliesTo("shared")) {
                    $userObject->addRole($roleObject);
                } else if ($roleObject->autoAppliesTo("standard")) {
                    $userObject->addRole($roleObject);
                }
            }
        }
    }

    /**
     * @todo MOVE TO BACKEND
     * @static
     * @param UserInterface $userObject
     */
    public static function updateAutoApplyRole(&$userObject)
    {
        return;
        $roles = self::getRolesList(array(), true);
        foreach ($roles as $roleObject) {
            if (!$userObject->canSee($roleObject)) continue;
            if ($roleObject->autoAppliesTo($userObject->getProfile()) || $roleObject->autoAppliesTo("all")) {
                $userObject->addRole($roleObject);
            }
        }
    }

    /**
     * @param UserInterface $userObject
     */
    public static function updateAuthProvidedData(&$userObject)
    {
        ConfService::getAuthDriverImpl()->updateUserObject($userObject);
    }

    /**
     * Retrieve the current users who have either read or write access to a repository
     * @param $repositoryId
     * @param string $rolePrefix
     * @param bool $splitByType
     * @return array
     */
    public static function getRolesForRepository($repositoryId, $rolePrefix = '', $splitByType = false)
    {
        $aclClient = MicroApi::GetAclServiceApi();
        $request = new RestSearchACLRequest();
        $repoQuery = new IdmACLSingleQuery();
        $repoQuery->setWorkspaceIDs([$repositoryId]);
        $read = new IdmACLAction();
        $read->setName("read");
        $write = new IdmACLAction();
        $write->setName("write");
        $repoQuery->setActions([$read, $write]);
        if($rolePrefix !== "") {
            $repoQuery->setRoleIDs([$rolePrefix . "*"]);
        }
        $request->setQueries([$repoQuery]);

        $result = [];
        $response = $aclClient->search($request);
        if($response->getAcLs() != null) {
            $acls = $response->getAcLs();
            foreach ($acls as $acl) {
                if ($acl->getRoleId() !== "" && !in_array($acl->getRoleId(), $result)){
                    $result[] = $acl->getRoleId();
                }
            }
        }

        if (!$splitByType) {
            return $result;
        }

        $rolesClient = MicroApi::GetRoleServiceApi();
        $request = new RestSearchRoleRequest();
        $query = new IdmRoleSingleQuery();
        $query->setUuid($result);
        $request->setQueries([$query]);
        $response = $rolesClient->search($request);
        $newResult = ["role_user" => [], "role_group" => [], "role_role" => []];
        if($response->getRoles() === null){
            return $newResult;
        }
        foreach($response->getRoles() as $role){
            if ($role->getGroupRole()) {
                $newResult["role_group"][] = $role;
            } else if ($role->getUserRole()){
                $newResult["role_user"][] = $role;
            } else {
                $newResult["role_role"][] = $role;
            }
        }
        return $newResult;

    }

    /**
     * @param $userId
     * @return Role[]
     */
    public static function getRolesOwnedBy($userId){

        $api = MicroApi::GetRoleServiceApi();
        $req = new RestSearchRoleRequest();
        $sQ = new IdmRoleSingleQuery();
        $sQ->setOwnerUuid($userId);
        $req->setQueries([$sQ]);
        $rolesList = $api->search($req);
        $result = [];
        if ($rolesList->getRoles() == null) {
            return $result;
        }
        foreach ($rolesList->getRoles() as $idmRole) {
            $roleId = $idmRole->getUuid();
            $result[$roleId] = new Role($roleId, $idmRole, true);
        }
        self::loadAcls($result);
        return $result;

    }

    /**
     * @param $roleId
     * @param $userId
     * @return Role
     */
    public static function getOrCreateOwnedRole($roleId, $userId){

        $existing = self::getOwnedRole($roleId, $userId);
        if (!empty($existing)) {
            return $existing;
        }
        $r = new Role($roleId);
        $r->setOwnerUuid($userId);
        self::updateRole($r);
        return $r;
    }

    /**
     * @param $roleId
     * @param $userId
     * @return Role
     */
    public static function getOwnedRole($roleId, $userId){

        $api = MicroApi::GetRoleServiceApi();
        $req = new RestSearchRoleRequest();
        $sQ = new IdmRoleSingleQuery();
        $sQ->setUuid([$roleId]);
        $sQ->setOwnerUuid($userId);
        $req->setQueries([$sQ]);
        $rolesList = $api->search($req);
        $result = [];
        if ($rolesList->getRoles() == null ){
            return null;
        }
        foreach ($rolesList->getRoles() as $idmRole) {
            if ($idmRole->getUuid() == $roleId){
                return new Role($roleId, $idmRole);
            }
        }
        return null;

    }

    /**
     * Get Role by Id
     *
     * @param string $roleId
     * @return Role|false
     */
    public static function getRole($roleId)
    {
        try{
            $existingRole = MicroApi::GetRoleServiceApi()->get($roleId);
        } catch (ApiException $ex){
            if($ex->getCode() == 404){
                return false;
            } else {
                throw $ex;
            }
        }
        return new Role($roleId, $existingRole);
    }

    /**
     * @param string $roleId Id of the role
     * @param string $groupPath GroupPath to be applied
     * @return Role
     */
    public static function getOrCreateRole($roleId, $groupPath = "/", $isGroupRole = false, $isUserRole = false)
    {
        $existingRole = self::getRole($roleId);
        if (!empty($existingRole)) {
            return $existingRole;
        }
        $role = new Role($roleId);
        $role->setGroupPath($groupPath);
        if($isGroupRole) {
            $role->setGroupRole(true);
        } else if ($isUserRole) {
            $role->setUserRole(true);
        }
        self::updateRole($role);
        return $role;
    }

    /**
     * Create or update role
     *
     * @param Role $roleObject
     * @param UserInterface|null $userObject
     */
    public static function updateRole($roleObject, $userObject = null)
    {
        $roleApi = MicroApi::GetRoleServiceApi();
        $label = $roleObject->getLabel();
        if (empty($label)) {
            if ($roleObject->getGroupRole()) {
                $roleObject->setLabel("Group " . $roleObject->getUuid());
            } else if ($roleObject->getUserRole()) {
                $roleObject->setLabel("User " . $roleObject->getUuid());
            } else {
                $roleObject->setLabel("Role " . $roleObject->getUuid());
            }
        }
        // Save Role
        $idmRole = $roleApi->set($roleObject->getUuid(), $roleObject);
        // Save Acls
        self::saveAcls($roleObject);
    }

    /**
     * Delete a role by its id
     * @static
     * @param string $roleId
     * @param string $ownerId
     * @return void
     * @throws PydioException
     */
    public static function deleteRole($roleId, $ownerId = null)
    {
        if(!empty($ownerId)){
            // Get Role
            $existingRole = MicroApi::GetRoleServiceApi()->get($roleId);
            if (!empty($existingRole) && $existingRole->getOwnerUuid() != $ownerId) {
                throw new PydioException("Cannot delete this role");
            }
        }
        MicroApi::GetRoleServiceApi()->delete($roleId);
    }

    /**
     * @param string $parentUserId
     * @return Role
     */
    public static function limitedRoleFromParent($parentUserId)
    {
        $parentRole = self::getRole($parentUserId);
        if ($parentRole === false) return null;

        $inheritActions = PluginsService::searchManifestsWithCache("//server_settings/param[@inherit='true']", function ($nodes) {
            $result = [];
            if (is_array($nodes) && count($nodes)) {
                foreach ($nodes as $node) {
                    $paramName = $node->getAttribute("name");
                    $pluginId = $node->parentNode->parentNode->getAttribute("id");
                    if (isSet($result[$pluginId])) $result[$pluginId] = array();
                    $result[$pluginId][] = $paramName;
                }
            }
            return $result;
        });

        // Clear ACL, Keep disabled actions, keep 'inherit' parameters.
        $childRole = new Role("PYDIO_PARENT_USR_/");
        $childRole->bunchUpdate(array(
            "ACL" => array(),
            "ACTIONS" => $parentRole->listAllActionsStates(),
            "APPLIES" => array(),
            "PARAMETERS" => array()));
        $params = $parentRole->listParameters();

        foreach ($params as $scope => $plugData) {
            foreach ($plugData as $pId => $paramData) {
                if (!isSet($inheritActions[$pId])) continue;
                foreach ($paramData as $pName => $pValue) {
                    $childRole->setParameterValue($pId, $pName, $pValue, $scope);
                }
            }
        }

        return $childRole;
    }

    /**
     * @param boolean $status
     */
    public static function enableRolesCache($status){
        self::$useCache = $status;
        if($status){
            self::$rolesCache = null;
        }
    }

    /** @var  boolean */
    private static $useCache;

    /** @var  array */
    private static $rolesCache;

    /**
     * Get all defined roles
     * @static
     * @param array $roleIds
     * @param boolean $excludeReserved,
     * @return Role[]
     */
    public static function getRolesList($roleIds = array(), $excludeReserved = false, $includeOwnedRoles = false)
    {
        $api = MicroApi::GetRoleServiceApi();
        $req = new RestSearchRoleRequest();
        $subQueries = [];
        if (count($roleIds) > 0) {
            $sQ = new IdmRoleSingleQuery();
            $sQ->setUuid($roleIds);
            $subQueries[] = $sQ;
        }
        if ($excludeReserved) {
            $sQ2 = new IdmRoleSingleQuery();
            $sQ2->setIsGroupRole(true);
            $sQ2->setIsUserRole(true);
            $sQ2->setNot(true);
            $subQueries[] = $sQ2;
        }
        if (!$includeOwnedRoles) {
            $sQ3 = new IdmRoleSingleQuery();
            $sQ3->setHasOwner(true);
            $sQ3->setNot(true);
            $subQueries[] = $sQ3;
        }
        $req->setQueries($subQueries);
        $req->setOperation(ServiceOperationType::_AND);
        $rolesList = $api->search($req);
        $roleData = [];
        if ($rolesList->getRoles() != null) {
            foreach($rolesList->getRoles() as $role) {
                $pydioRole = new Role($role->getUuid(), $role, true);
                $roleData[$role->getUuid()] = $pydioRole;
            }
        }
        self::loadAcls($roleData);
        return $roleData;

    }

    /**
     * @param $roles Role[]
     */
    public static function loadAcls(&$roles) {

        $wsIds = [];
        foreach($roles as $role){
            $wsIds[$role->getId()] = [];
            $role->setAcl("pydiogateway", "rw");
        }
        // List ACLS
        $api = MicroApi::GetAclServiceApi();
        $request = new RestSearchACLRequest();
        $query = new IdmACLSingleQuery();
        $query->setRoleIDs(array_keys($roles));
        $request->setQueries([$query]);
        $collection = $api->search($request);
        if ($collection->getAcLs() == null) {
            return;
        }

        $acls = $collection->getAcLs();
        foreach($acls as $acl) {

            $actionName = $acl->getAction()->getName();
            $workspaceId = $acl->getWorkspaceId();
            $roleId = $acl->getRoleId();
            if(!isSet($roles[$roleId])) {
                continue;
            }
            $role = &$roles[$roleId];

            if($actionName == "read" || $actionName == "write" || $actionName == "deny") {
                if (!isSet($wsIds[$roleId][$workspaceId])) $wsIds[$roleId][$workspaceId] = ["read"=>false, "write"=>false, "deny" => false];
                $wsIds[$roleId][$workspaceId][$actionName] = true;
            } else if(strpos($actionName, "action:") === 0) {
                list($a, $pluginId, $realActionName) = explode(":", $actionName);
                $actionValue = $acl->getAction()->getValue();
                $role->setActionState($pluginId, $realActionName, $workspaceId, json_decode($actionValue, true));
            } else if(strpos($actionName, "parameter:") === 0) {
                list($a, $pluginId, $paramName) = explode(":", $actionName);
                $paramValue = $acl->getAction()->getValue();
                $role->setParameterValue($pluginId, $paramName, json_decode($paramValue), $workspaceId);
            }
        }

        foreach($wsIds as $roleId => $wsRights) {
            if(!isSet($roles[$roleId])) {
                continue;
            }
            $role = &$roles[$roleId];
            foreach($wsRights as $wsId => $rights) {
                if ($rights["deny"]) {
                    $role->setAcl($wsId, PYDIO_VALUE_CLEAR);
                } else if ($rights["read"] || $rights["deny"]) {
                    $role->setAcl($wsId, ($rights["read"] ? "r" : "") . ($rights["write"] ? "w" : ""));
                }
            }
        }

    }

    protected static function saveAcls(Role $roleObject) {

        $roleID = $roleObject->getId();

        $api = MicroApi::GetAclServiceApi();
        $idmAcl = new IdmACL();
        $idmAcl->setRoleId($roleID);
        $api->delete($idmAcl);

        $acls = [];

        $roleAcls = $roleObject->listAcls(false);
        foreach ($roleAcls as $workspaceId => $acl) {

            $workspaceId .= ""; // make sure it's a string
            $workspace = RepositoryService::getRepositoryById($workspaceId);

            if (empty($workspace)) {
                continue;
            }
            if ($workspace->isWriteable()){
                $rootNodes = $workspace->getRootNodes();
            } else {
                // This is a special ws
                $root = new RepositoryRoot();
                $root->setUuid("$workspaceId-ROOT");
                $rootNodes = [$root];
            }
            foreach ($rootNodes as $rootNode) {
                $nodeId = $rootNode->getUuid();
                // Removing the acls and adding the Deny
                if ($acl === PYDIO_VALUE_CLEAR) {
                    $acls[] = [$workspaceId, $nodeId, "deny", "1"];
                    continue;
                }

                $read = strpos($acl, "r");
                $write = strpos($acl, "w");

                if ($read !== FALSE) {
                    $acls[] = [$workspaceId, $nodeId, "read", "1"];
                }

                if ($write !== FALSE) {
                    $acls[] = [$workspaceId, $nodeId, "write", "1"];
                }
            }

        }

        $actions = $roleObject->listAllActionsStates();
        foreach ($actions as $workspaceId => $pluginData) {
            foreach ($pluginData as $pluginId => $actions) {
                foreach($actions as $actionName => $actionValue) {
                    $acls[] = [$workspaceId, "", "action:$pluginId:$actionName", json_encode($actionValue)];
                }
            }
        }

        $parameters = $roleObject->listParameters(true);
        foreach ($parameters as $workspaceId => $pluginData) {
            foreach ($pluginData as $pluginId => $params) {
                foreach ($params as $paramName => $paramValue) {
                    $acls[] = [$workspaceId, "", "parameter:$pluginId:$paramName", json_encode($paramValue)];
                }
            }
        }

        foreach($acls as $acl) {

            $idmAcl = new IdmACL();
            $action = new IdmACLAction();
            $idmAcl->setRoleId($roleID);
            $idmAcl->setWorkspaceId($acl[0]);
            $idmAcl->setNodeId($acl[1]);
            $action->setName($acl[2]);
            $action->setValue($acl[3]);
            $idmAcl->setAction($action);

            $api->put($idmAcl);

        }

    }



    /**
     * Specific operations to perform at boot time
     * @static
     * @throws PydioException
     * @throws \Exception
     */
    public static function bootSequence()
    {
        $bootConfDir = PYDIO_DATA_PATH. "/plugins/boot.conf";
        if (file_exists(PYDIO_CACHE_DIR . "/admin_counted") || file_exists($bootConfDir."/admin_counted")) return;
        try{
            $rootRole = RolesService::getOrCreateRole(RolesService::RootGroup, "/", true);
            $rootRole->setLabel("Root Group");
            $allRepos = RepositoryService::listAllRepositories();
            foreach ($allRepos as $repositoryId => $repoObject) {
                if ($repoObject->isTemplate) continue;
                $gp = $repoObject->getGroupPath();
                if (empty($gp) || $gp == "/") {
                    if ($repoObject->getDefaultRight() != "") {
                        $rootRole->setAcl($repositoryId, $repoObject->getDefaultRight());
                    }
                }
            }
            $parameters = PluginsService::searchManifestsWithCache("//server_settings/param[@scope]", function ($paramNodes) {
                $result = [];
                /** @var \DOMElement $xmlNode */
                foreach ($paramNodes as $xmlNode) {
                    $default = $xmlNode->getAttribute("default");
                    if (empty($default)) continue;
                    $parentNode = $xmlNode->parentNode->parentNode;
                    $pluginId = $parentNode->getAttribute("id");
                    if (empty($pluginId)) {
                        $pluginId = $parentNode->nodeName . "." . $parentNode->getAttribute("name");
                    }
                    $result[] = ["pluginId" => $pluginId, "name" => $xmlNode->getAttribute("name"), "default" => $default];
                }
                return $result;
            });
            foreach ($parameters as $parameter) {
                $rootRole->setParameterValue($parameter["pluginId"], $parameter["name"], $parameter["default"]);
            }
            RolesService::updateRole($rootRole);

            $miniRole = RolesService::getRole("EXTERNAL_USERS");
            if ($miniRole === false) {
                $rootRole = new Role("EXTERNAL_USERS");
                $rootRole->setLabel("External users");
                $actions = array(
                    "action.share" => ["open_user_shares"],
                    "action.user" => ["open_address_book"]
                );
                $rootRole->setAutoApplies(array("shared"));
                foreach ($actions as $pluginId => $acts) {
                    foreach ($acts as $act) {
                        $rootRole->setActionState($pluginId, $act, PYDIO_REPO_SCOPE_ALL, false);
                    }
                }
                RolesService::updateRole($rootRole);
            }
            $miniRole = RolesService::getRole("MINISITE");
            if ($miniRole === false) {
                $rootRole = new Role("MINISITE");
                $rootRole->setLabel("Minisite Users");
                $actions = array(
                    "meta.watch" => ["toggle_watch"],
                    "conf.serial" => ["get_bookmarks"],
                    "conf.sql" => ["get_bookmarks"],
                    "action.share" => ["share", "share_react", "share-edit-shared", "share-folder-workspace", "share-file-minisite", "share-selection-minisite", "share-folder-minisite-public", "open_user_shares"],
                    "gui.ajax" => ["bookmark"],
                    "auth.sql" => ["pass_change"],
                    "action.user" => ["open_address_book"]
                );
                foreach ($actions as $pluginId => $acts) {
                    foreach ($acts as $act) {
                        $rootRole->setActionState($pluginId, $act, PYDIO_REPO_SCOPE_SHARED, false);
                    }
                }
                RolesService::updateRole($rootRole);
            }
            $miniRole = RolesService::getRole("MINISITE_NODOWNLOAD");
            if ($miniRole === false) {
                $rootRole = new Role("MINISITE_NODOWNLOAD");
                $rootRole->setLabel("Minisite Users - No Download");
                $actions = array(
                    "access.fs" => array("download", "download_chunk", "prepare_chunk_dl", "download_all")
                );
                foreach ($actions as $pluginId => $acts) {
                    foreach ($acts as $act) {
                        $rootRole->setActionState($pluginId, $act, PYDIO_REPO_SCOPE_SHARED, false);
                    }
                }
                RolesService::updateRole($rootRole);
            }
            $miniRole = RolesService::getRole("GUEST");
            if ($miniRole === false) {
                $rootRole = new Role("GUEST");
                $rootRole->setLabel("Guest user role");
                $actions = array(
                    "access.fs" => array("purge"),
                    "meta.watch" => array("toggle_watch"),
                    "index.lucene" => array("index"),
                );
                $rootRole->setAutoApplies(array("guest"));
                foreach ($actions as $pluginId => $acts) {
                    foreach ($acts as $act) {
                        $rootRole->setActionState($pluginId, $act, PYDIO_REPO_SCOPE_ALL);
                    }
                }
                RolesService::updateRole($rootRole);
            }
            if(!file_exists($bootConfDir)){
                mkdir($bootConfDir, 0666, true);
            }
            file_put_contents($bootConfDir . "/admin_counted", "true");
        } catch (ApiException $e){
            // Cannot contact API, just ignore
        }

    }
}
