<?php
/*
 * Copyright 2007-2017 Abstrium <contact (at) pydio.com>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com>.
 */
namespace Pydio\OCS\Model;

use Pydio\Core\Http\Client\SimpleStoreApi;
use Pydio\Core\Services\ConfService;
use Pydio\Conf\Sql\SqlConfDriver;
use Swagger\Client\ApiException;

defined('PYDIO_EXEC') or die('Access not allowed');

define('OCS_SQLSTORE_FORMAT', 'serial');
define('OCS_SQLSTORE_NS_INVITATION', 'ocs_invitation');
define('OCS_SQLSTORE_NS_REMOTE_SHARE', 'ocs_remote_share');

/**
 * Class SQLStore
 * @package Pydio\OCS\Model
 */
class SQLStore implements IStore
{
    /**
     * @var SimpleStoreApi
     */
    protected $api;

    /**
     * SQLStore constructor.
     */
    public function __construct()
    {
        $this->api = new SimpleStoreApi();
    }

    /**
     * @param ShareInvitation $invitation
     * @return int
     */
    public function generateInvitationId(ShareInvitation &$invitation){
        if(empty($id)){
            $id = $this->findAvailableID(OCS_SQLSTORE_NS_INVITATION);
            $invitation->setId($id);
        }
        return $invitation->getId();
    }

    /**
     * Persists an invitation to store
     * @param ShareInvitation $invitation
     * @return ShareInvitation|false The invitation with its new ID.
     */
    public function storeInvitation(ShareInvitation $invitation)
    {
        $id = $invitation->getId();
        $data = [
            "INVITATION" => serialize($invitation),
        ];
        $indexableData = [
            "LINK_HASH" => $invitation->getLinkHash(),
        ];
        $this->api->storeDocument(OCS_SQLSTORE_NS_INVITATION, $id, "", $data, $indexableData);
        return $invitation;
    }

    /**
     * Find all invitatins for a given token
     * @param string $linkToken
     * @return ShareInvitation[]
     */
    public function invitationsForLink($linkToken)
    {
        $cursor = null;
        $metaQuery = "+LINK_HASH:" . SimpleStoreApi::escapeMetaValue($linkToken);
        $docs = $this->api->listDocuments(OCS_SQLSTORE_NS_INVITATION, $metaQuery);
        $result = [];
        foreach($docs as $doc){
            $result[] = unserialize($doc["INVITATION"]);
        }
        return $result;
    }

    /**
     * Find an invitation by ID
     * @param $invitationId
     * @return ShareInvitation|null
     */
    public function invitationById($invitationId)
    {
        $doc = $this->api->loadDocument(OCS_SQLSTORE_NS_INVITATION, $invitationId);
        return unserialize($doc["INVITATION"]);
    }

    /**
     * Delete an invitation
     * @param ShareInvitation $invitation
     * @return bool
     */
    public function deleteInvitation(ShareInvitation $invitation)
    {
        return $this->api->deleteDocuments(OCS_SQLSTORE_NS_INVITATION, $invitation->getId());
    }

    /**
     * Persists a remote share to the store
     * @param RemoteShare $remoteShare
     * @return RemoteShare|false The share with eventually its new ID.
     */
    public function storeRemoteShare(RemoteShare $remoteShare)
    {
        $id = $remoteShare->getId();
        if(empty($id)){
            $id = $remoteShare->getOcsToken();
            $remoteShare->setId($id);
        }
        $data = [
            "REMOTE_SHARE" => serialize($remoteShare)
        ];
        $indexableData = [
            "SHARE_OWNER" => $remoteShare->getUser(),
            "OCS_REMOTE_ID" => $remoteShare->getOcsRemoteId()
        ];
        $this->api->storeDocument(OCS_SQLSTORE_NS_REMOTE_SHARE, $id, $remoteShare->getUser(), $data, $indexableData);
        return $remoteShare;

    }

    /**
     * Find all remote shares for a given user
     * @param string $userName
     * @return RemoteShare[]
     */
    public function remoteSharesForUser($userName)
    {
        $metaQuery = "+SHARE_OWNER:" . SimpleStoreApi::escapeMetaValue($userName);
        try{
            $docs = $this->api->listDocuments(OCS_SQLSTORE_NS_REMOTE_SHARE, $metaQuery);
        } catch(Exception $e){
            return [];
        }
        $res = [];
        foreach($docs as $doc) {
            $res[] = unserialize($doc["REMOTE_SHARE"]);
        }
        return $res;
    }

    /**
     * Find a remote share by its id
     * @param string $remoteShareId
     * @return RemoteShare|false
     */
    public function remoteShareById($remoteShareId)
    {
        $doc = $this->api->loadDocument(OCS_SQLSTORE_NS_REMOTE_SHARE, $remoteShareId);
        return unserialize($doc["REMOTE_SHARE"]);
    }

    /**
     * @param $ocsRemoteId
     * @return mixed|null
     */
    public function remoteShareForOcsRemoteId($ocsRemoteId){
        $metaQuery = "+OCS_REMOTE_ID:" . SimpleStoreApi::escapeMetaValue($ocsRemoteId);
        $docs = $this->api->listDocuments(OCS_SQLSTORE_NS_REMOTE_SHARE, $metaQuery);
        $res = [];
        foreach($docs as $doc) {
            $res[] = unserialize($doc["REMOTE_SHARE"]);
            break;
        }
        if (count($res)) {
            return $res[0];
        } else {
            return null;
        }
    }

    /**
     * Delete an existing remote share
     * @param RemoteShare $remoteShare
     * @return bool
     */
    public function deleteRemoteShare(RemoteShare $remoteShare)
    {
        $this->api->deleteDocuments(OCS_SQLSTORE_NS_REMOTE_SHARE, $remoteShare->getId());
        return true;
    }

    /**
     * @param $namespace
     * @return int
     */
    protected function findAvailableID($namespace){
        $id = 0;
        while(true){
            $id = rand();
            $data = $this->api->loadDocument($namespace, $id);
            if(empty($data)) {
                break;
            }
        }
        return $id;
    }

    /**
     * @return string
     */
    protected function getGUID(){
        if (function_exists('com_create_guid')){
            return com_create_guid();
        }else{
            mt_srand((double)microtime()*10000);//optional for php 4.2.0 and up.
            $charid = strtoupper(md5(uniqid(rand(), true)));
            $hyphen = chr(45);// "-"
            $uuid = ""//chr(123)// "{"
                .substr($charid, 0, 8).$hyphen
                .substr($charid, 8, 4).$hyphen
                .substr($charid,12, 4).$hyphen
                .substr($charid,16, 4).$hyphen
                .substr($charid,20,12);
                //.chr(125);// "}"
            return $uuid;
        }
    }

}