/*
 * Copyright 2007-2017 Charles du Jeu - Abstrium SAS <team (at) pyd.io>
 * This file is part of Pydio.
 *
 * Pydio is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pydio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Pydio.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The latest code can be found at <https://pydio.com>.
 */

'use strict';

exports.__esModule = true;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var _modelAjxpNode = require('../model/AjxpNode');

var _modelAjxpNode2 = _interopRequireDefault(_modelAjxpNode);

var _utilPathUtils = require('../util/PathUtils');

var _utilPathUtils2 = _interopRequireDefault(_utilPathUtils);

var _lodashDebounce = require('lodash.debounce');

var _lodashDebounce2 = _interopRequireDefault(_lodashDebounce);

/**
 * WebSocket client
 */

var ReconnectingWebSocket = require('reconnecting-websocket');

var PydioWebSocket = (function () {

    /**
     *
     * @param pydioObject Pydio
     */

    function PydioWebSocket(pydioObject) {
        var _this = this;

        _classCallCheck(this, PydioWebSocket);

        this.pydio = pydioObject;

        this.currentRepo = null;
        this.ws = null;
        this.reloadRepositoriesDebounced = _lodashDebounce2['default'](function () {
            _this.pydio.reloadRepositoriesList();
        }, 500);

        this.pydio.observe("repository_list_refreshed", (function (data) {

            var repoId = undefined;
            if (data.active) {
                repoId = data.active;
            } else if (this.pydio.repositoryId) {
                repoId = this.pydio.repositoryId;
            }
            if (this.currentRepo && this.currentRepo == repoId) {
                // Ignore, repoId did not change!
                return;
            }
            this.currentRepo = repoId;
            this.open();
        }).bind(this));
    }

    PydioWebSocket.prototype.close = function close() {
        if (this.ws === null) {
            return;
        }
        this.ws.close(1000, 'Closing', { keepClosed: true });
    };

    /**
     * Open a connection
     */

    PydioWebSocket.prototype.open = function open() {
        var _this2 = this;

        this.close();
        if (!this.currentRepo) {
            return;
        }
        var url = this.pydio.getPluginConfigs("core.conf").get("ENDPOINT_WEBSOCKET");
        this.ws = new ReconnectingWebSocket(url);

        this.ws.addEventListener('open', function () {
            PydioWebSocket.subscribeJWT(_this2.ws);
        });
        this.ws.addEventListener('message', this.parseWebsocketMessage.bind(this));
        this.ws.addEventListener('close', function (event) {
            PydioWebSocket.logClose(event);
        });
    };

    PydioWebSocket.prototype.parseWebsocketMessage = function parseWebsocketMessage(msg) {
        var dm = this.pydio.getContextHolder();
        var event = JSON.parse(msg.data);
        if (event === "dump") {
            return;
        }
        if (event['@type']) {
            this.pydio.fire("websocket_event:" + event['@type'], event);
        }
        if (event['@type'] && event['@type'] === "idm") {
            this.reloadRepositoriesDebounced();
            return;
        }
        if (event['TaskUpdated'] && event['Job']) {
            this.pydio.fire("task_message", event);
            return;
        }
        if (!event.Type && event.Target) {
            event.Type = "CREATE";
        }
        var target = undefined;
        var currentRepoId = this.pydio.user.getActiveRepository();
        if (!this.pydio.user.repositories.has(currentRepoId)) {
            return;
        }
        var currentRepoSlug = this.pydio.user.repositories.get(currentRepoId).getSlug();
        switch (event.Type) {
            case "CREATE":
                target = PydioWebSocket.parseJSONNode(event.Target, currentRepoId, currentRepoSlug);
                if (target === null) {
                    return;
                }
                console.log("CREATE", target);
                dm.addNode(target, true);
                break;
            case "UPDATE_PATH":
            case "UPDATE_META":
            case "UPDATE_CONTENT":
                target = PydioWebSocket.parseJSONNode(event.Target, currentRepoId, currentRepoSlug);
                if (target === null) {
                    return;
                }
                if (event.Source) {
                    var _source = PydioWebSocket.parseJSONNode(event.Source, currentRepoId, currentRepoSlug);
                    target.getMetadata().set("original_path", _source.getPath());
                } else {
                    target.getMetadata().set("original_path", target.getPath());
                }
                console.log("UPDATE", target);
                dm.updateNode(target, true);
                break;
            case "DELETE":
                var source = PydioWebSocket.parseJSONNode(event.Source, currentRepoId, currentRepoSlug);
                if (source === null) {
                    return;
                }
                console.log("DELETE", source.getPath());
                dm.removeNodeByPath('/' + source.getPath());
                break;
            default:
                break;
        }
    };

    /**
     * @return AjxpNode | null
     * @param obj
     * @param currentRepoId string
     * @param currentRepoSlug string
     */

    PydioWebSocket.parseJSONNode = function parseJSONNode(obj, currentRepoId, currentRepoSlug) {

        var wsId = JSON.parse(obj.MetaStore.EventWorkspaceId);
        var nodeName = undefined;
        if (obj.MetaStore.name) {
            nodeName = JSON.parse(obj.MetaStore.name);
        } else {
            nodeName = _utilPathUtils2['default'].getBasename(obj.Path);
        }
        if (wsId == currentRepoId) {
            obj.Path = obj.Path.substr(currentRepoSlug.length + 1);
        } else {
            return null;
        }

        var node = new _modelAjxpNode2['default']('/' + obj.Path, obj.Type === "LEAF", nodeName, '', null);

        var meta = obj.MetaStore;
        for (var k in meta) {
            if (meta.hasOwnProperty(k)) {
                var metaValue = JSON.parse(meta[k]);
                node.getMetadata().set(k, metaValue);
                if (typeof metaValue == 'object') {
                    for (var kSub in metaValue) {
                        if (metaValue.hasOwnProperty(kSub)) {
                            node.getMetadata().set(kSub, metaValue[kSub]);
                        }
                    }
                }
            }
        }
        node.getMetadata().set('filename', node.getPath());
        if (obj.Size !== undefined) {
            node.getMetadata().set('bytesize', obj.Size);
        }
        if (obj.MTime !== undefined) {
            node.getMetadata().set('ajxp_modiftime', obj.MTime);
        }
        if (obj.Etag !== undefined) {
            node.getMetadata().set('etag', obj.Etag);
        }
        if (obj.Uuid !== undefined) {
            node.getMetadata().set('uuid', obj.Uuid);
        }
        return node;
    };

    PydioWebSocket.subscribeJWT = function subscribeJWT(ws) {
        PydioApi.getClient().request({ get_action: 'jwt' }, function (t) {
            var resp = t.responseJSON;
            if (resp !== null && resp['jwt']) {
                ws.send(JSON.stringify({
                    "@type": "subscribe",
                    "jwt": resp['jwt']
                }));
            }
        });
    };

    PydioWebSocket.logClose = function logClose(event) {
        var reason = undefined;
        switch (event.code) {
            case 1000:
                reason = "Normal closure, meaning that the purpose for which the connection was established has been fulfilled.";
                break;
            case 1001:
                reason = "An endpoint is \"going away\", such as a server going down or a browser having navigated away from a page.";
                break;
            case 1002:
                reason = "An endpoint is terminating the connection due to a protocol error";
                break;
            case 1003:
                reason = "An endpoint is terminating the connection because it has received a type of data it cannot accept (e.g., an endpoint that understands only text data MAY send this if it receives a binary message).";
                break;
            case 1004:
                reason = "Reserved. The specific meaning might be defined in the future.";
                break;
            case 1005:
                reason = "No status code was actually present.";
                break;
            case 1006:
                reason = "The connection was closed abnormally, e.g., without sending or receiving a Close control frame";
                break;
            case 1007:
                reason = "An endpoint is terminating the connection because it has received data within a message that was not consistent with the type of the message (e.g., non-UTF-8 [http://tools.ietf.org/html/rfc3629] data within a text message).";
                break;
            case 1008:
                reason = "An endpoint is terminating the connection because it has received a message that \"violates its policy\". This reason is given either if there is no other sutible reason, or if there is a need to hide specific details about the policy.";
                break;
            case 1009:
                reason = "An endpoint is terminating the connection because it has received a message that is too big for it to process.";
                break;
            case 1010:
                // Note that this status code is not used by the server, because it can fail the WebSocket handshake instead.
                reason = "An endpoint (client) is terminating the connection because it has expected the server to negotiate one or more extension, but the server didn't return them in the response message of the WebSocket handshake. Specifically, the extensions that are needed are: " + event.reason;
                break;
            case 1011:
                reason = "A server is terminating the connection because it encountered an unexpected condition that prevented it from fulfilling the request.";
                break;
            case 1015:
                reason = "The connection was closed due to a failure to perform a TLS handshake (e.g., the server certificate can't be verified).";
                break;
            default:
                reason = "Unknown reason";
                break;
        }
        if (console) {
            console.error("WebSocket Closed Connection for this reason :" + reason + " (code " + event.code + ")");
        }
    };

    return PydioWebSocket;
})();

exports['default'] = PydioWebSocket;
module.exports = exports['default'];
